open Printf

type compil_unit = {
  package : path option;
  imports : import list;
  class_decls : class_decl list;
  interf_decls : interf_decl list;
}

and class_decl = {
  class_modifiers : modifier list;
  class_name : string;
  class_extend : class_type;
  class_implements : class_type list;
  class_fields : field list;
  class_methods : method_def list;
  class_initializers : block list;
  class_constructors : constructor list;
}

and interf_decl = {
  interf_modifiers : modifier list;
  interf_name : string;
  interf_extends : class_type list;
  interf_constants : field list;
  interf_abstracts : abstract_decl list;
}

(* class members *)

and field = {
  field_modifiers : modifier list;
  field_type : java_type;
  field_name : string;
  field_initializer : init_expr option;
}

and method_def = {
  method_modifiers : modifier list;
  method_return : java_type option;
  method_name : string;
  method_parameters : (java_type * string) list;
  method_throws : class_type list;
  method_body : block option;
}

and constructor = {
  constructor_modifiers : modifier list;
  constructor_name : string;
  constructor_parameters : (java_type * string) list;
  constructor_throws : class_type list;
  constructor_call : ccall;
  constructor_body : block option;
}

(* interface members *)

and abstract_decl = {
  abstract_modifiers : modifier list;
  abstract_return : java_type option;
  abstract_name : string;
  abstract_parameters : (java_type * string) list;
  abstract_throws : class_type list;
}

(* modifiers *)

and modifier = [
  `abstract | `final | `native | `private_ | `protected | `public | `static |
      `synchronized | `transient | `volatile
]

(* tools *)

and path = string list

and import = Isingle of class_type | Iall of path

and ccall = Cnone | Cthis of expression list | Csuper of expression list

and init_expr = Simple_init of expression | Array_init of init_expr list

and switch_label = Scase of expression | Sdefault

and for_init =
  | Fnone
  | Fexprs of expression list
  | Fdecls of (java_type * string * init_expr option) list

and affect =
  | Aname of string list (* truc.bidule.truc *)
  | Afield of expression * string (* getObj().truc = *)
  | Aarray of expression * expression (* getArray()[3] *)

(* types *)

and class_type = string list

and java_type = [ `boolean 
| `byte | `char | `short | `int | `long 
| `float | `double 
| `string
| `java_object of class_type | `array of java_type 
| `void | `nil ]

(* instructions *)

and block = instruction list

and instruction =
  | Ilabeled of string * instruction
  | Iexpr of expression
  | Idecl of java_type * string * init_expr option
  | Iblock of instruction list
  | Iifthenelse of expression * instruction * instruction
  | Iswitch of expression * (switch_label list * block) list
  | Iwhile of expression * instruction
  | Idowhile of instruction * expression
  | Ifor of for_init * expression option * expression list * instruction
  | Ibreak of string option
  | Icontinue of string option
  | Ireturn of expression option
  | Ithrow of expression
  | Isynchronized of expression * block
  | Itry of block * (java_type * string * block) list * block option
  | Inop
  | Iprintln of expression

(* expressions *)

and expression =
  | Eunary of (unop * expression)
  | Ebinary of (binop * expression * expression)
  | Ebincomp of (bincomp * expression * expression) (* compare... *)
  | Eternary of (terop * expression * expression * expression)
  | Ebool of bool
  | Ebyte of int
  | Echar of int
  | Eshort of int
  | Eint of int
  | Efloat of float
  | Edouble of float
  | Estring of string
  | Enew_object of class_type * expression list
  | Enew_array of java_type * expression list * int
  | Efield_access of expression * string
  | Ecall of string * expression list
  | Eexpr_call of expression * string * expression list
  | Earray_access of expression * expression
  | Etype_cast of java_type * expression
  | Eexpr_cast of expression * expression
  | Einstanceof of expression * java_type
  | Eaffect of affect * expression
  | Evar of string
  | Enull
  | Ethis
  | Esuper

and unop = 
| Oneg | Onot | Otild
| Opreinc | Opredec | Opostinc | Opostdec

and binop = Oplus  (* int, float, string *)
| Ominus | Omul | Odiv  (* int or float *)
| Omod | Oor | Oand | Oxor | Oinfinf | Osupsup | Osupsupsup  (* int (and bool for AND, OR, XOR) *)
| Oandand | Ooror  (* bool *)

and bincomp = Oinf | Oinfequal | Osup | Osupequal | Oequal | Onotequal

and terop = Oifthenelse
;;

let rec string_of_java_type jt =
  match jt with
  | `boolean -> "boolean"
  | `byte -> "byte"
  | `char -> "char"
  | `short -> "short"
  | `int -> "int"
  | `long -> "long"
  | `float -> "float"
  | `double -> "double"
  | `string -> "string"
  | `void -> "void"
  | `nil -> "null"
  | `java_object(o) -> (match o with [] -> "object" | h::t -> h)
  | `array(t) -> (string_of_java_type t) ^ "[]";;

let string_of_java_types l = 
  let rec sojt l = 
  match l with 
    | [] -> ""
    | h::[] -> (string_of_java_type h)
    | h::t -> (string_of_java_type h) ^ ", " ^ (sojt t)
in (sojt l);;

let string_of_unary_op op =
  match op with
  | Oneg -> "-"
  | Onot -> "!"
  | Otild -> "~"
  | Opreinc | Opostinc -> "++"
  | Opredec | Opostdec -> "--";;

let string_of_binary_op op =
  match op with
  | Oplus -> "+"
  | Ominus -> "-"
  | Omul -> "*"
  | Odiv -> "/"
  | Omod -> "%"
  | Oor -> "|"
  | Oand -> "&"
  | Oxor -> "^"
  | Oinfinf -> "<<"
  | Osupsup -> ">>"
  | Osupsupsup -> ">>>"
  | Oandand -> "&&"
  | Ooror -> "||";;

let string_of_binary_comp op = 
  match op with
    | Oinf -> "<"
    | Oinfequal -> "<="
    | Osup -> ">"
    | Osupequal -> ">="
    | Oequal -> "="
    | Onotequal -> "<>";;


